# test_utils.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pytest

from pomegranate._utils import _cast_as_tensor
from pomegranate._utils import _update_parameter
from pomegranate._utils import _check_parameter
from pomegranate._utils import partition_sequences

from .tools import assert_almost_equal
from .tools import assert_equal
from .tools import assert_raises
from numpy.testing import assert_array_equal
from numpy.testing import assert_array_almost_equal


def _test_cast(y, dtype, ndim):
	x = _cast_as_tensor(y)
	assert isinstance(x, torch.Tensor)
	assert x.dtype == dtype
	assert x.ndim == ndim

	if type(x) not in (numpy.ndarray, torch.Tensor) and ndim == 0:
		if type(x) in float:
			assert_almost_equal(x, y)
		else:
			assert_equal(x, y)
	else:
		assert_array_almost_equal(x, y)


def test_cast_as_tensor_bool():
	_test_cast(False, torch.bool, 0)
	_test_cast(True, torch.bool, 0)


def test_cast_as_tensor_int():
	_test_cast(5, torch.int64, 0)
	_test_cast(0, torch.int64, 0)
	_test_cast(-1, torch.int64, 0)


def test_cast_as_tensor_float():
	_test_cast(1.2, torch.float32, 0)
	_test_cast(0.0, torch.float32, 0)
	_test_cast(-8.772, torch.float32, 0)


def test_cast_as_tensor_numpy_bool():
	_test_cast(numpy.array(False), torch.bool, 0)
	_test_cast(numpy.array(True), torch.bool, 0)


def test_cast_as_tensor_numpy_int():
	_test_cast(numpy.array(5), torch.int64, 0)
	_test_cast(numpy.array(0), torch.int64, 0)
	_test_cast(numpy.array(-1), torch.int64, 0)


def test_cast_as_tensor_numpy_float():
	_test_cast(numpy.array(1.2), torch.float64, 0)
	_test_cast(numpy.array(0.0), torch.float64, 0)
	_test_cast(numpy.array(-8.772), torch.float64, 0)


def test_cast_as_tensor_torch_bool():
	_test_cast(torch.tensor(False), torch.bool, 0)
	_test_cast(torch.tensor(True), torch.bool, 0)


def test_cast_as_tensor_torch_int():
	_test_cast(torch.tensor(5), torch.int64, 0)
	_test_cast(torch.tensor(0), torch.int64, 0)
	_test_cast(torch.tensor(-1), torch.int64, 0)


def test_cast_as_tensor_torch_float():
	_test_cast(torch.tensor(1.2), torch.float32, 0)
	_test_cast(torch.tensor(0.0), torch.float32, 0)
	_test_cast(torch.tensor(-8.772), torch.float32, 0)


def test_cast_as_tensor_check_wrong():
	assert_raises(AssertionError, _test_cast, True, torch.int64, 0)
	assert_raises(AssertionError, _test_cast, True, torch.bool, 1)
	assert_raises(AssertionError, _test_cast, 1, torch.int32, 0)
	assert_raises(AssertionError, _test_cast, 1, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, 1.2, torch.int64, 0)
	assert_raises(AssertionError, _test_cast, 1.2, torch.float64, 1)



def test_cast_as_tensor_numpy_bool_1d():
	_test_cast(numpy.array([True, False, True, True]), torch.bool, 1)
	_test_cast(numpy.array([True, True, True]), torch.bool, 1)
	_test_cast(numpy.array([False]), torch.bool, 1)


def test_cast_as_tensor_numpy_int_1d():
	_numpy_dtypes = numpy.int32, numpy.int64
	_torch_dtypes = torch.int32, torch.int64

	for _dtype1, _dtype2 in zip(_numpy_dtypes, _torch_dtypes):
		_test_cast(numpy.array([1, 2, 3], dtype=_dtype1), _dtype2, 1)
		_test_cast(numpy.array([0, -3, 0], dtype=_dtype1), _dtype2, 1)
		_test_cast(numpy.array([0], dtype=_dtype1), _dtype2, 1)


def test_cast_as_tensor_numpy_float_1d():
	_numpy_dtypes = numpy.float16, numpy.float32, numpy.float64
	_torch_dtypes = torch.float16, torch.float32, torch.float64

	for _dtype1, _dtype2 in zip(_numpy_dtypes, _torch_dtypes):
		_test_cast(numpy.array([1.2, 2.0, 3.1], dtype=_dtype1), _dtype2, 1)
		_test_cast(numpy.array([0.0, -3.0, 0.0], dtype=_dtype1), _dtype2, 1)
		_test_cast(numpy.array([0.0], dtype=_dtype1), _dtype2, 1)


def test_cast_as_tensor_numpy_check_wrong_1d():
	x1 = numpy.array([True, True, False])
	x2 = numpy.array([1, 2, 3])
	x3 = numpy.array([1.0, 1.1, 1.1])

	assert_raises(AssertionError, _test_cast, x1, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, x1, torch.bool, 0)
	assert_raises(AssertionError, _test_cast, x2, torch.int32, 1)
	assert_raises(AssertionError, _test_cast, x2, torch.int64, 0)
	assert_raises(AssertionError, _test_cast, x3, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, x3, torch.float64, 0)



def test_cast_as_tensor_numpy_bool_2d():
	_test_cast(numpy.array([[True, False, True], [True, True, False]]), 
		torch.bool, 2)
	_test_cast(numpy.array([[True, True, True]]), torch.bool, 2)
	_test_cast(numpy.array([[False]]), torch.bool, 2)


def test_cast_as_tensor_numpy_int_2d():
	_numpy_dtypes = numpy.int32, numpy.int64
	_torch_dtypes = torch.int32, torch.int64

	for _dtype1, _dtype2 in zip(_numpy_dtypes, _torch_dtypes):
		_test_cast(numpy.array([[1, 2, 3], [3, 2, 1]], dtype=_dtype1), 
			_dtype2, 2)
		_test_cast(numpy.array([[0, -3, 0]], dtype=_dtype1), _dtype2, 2)
		_test_cast(numpy.array([[0]], dtype=_dtype1), _dtype2, 2)


def test_cast_as_tensor_numpy_float_2d():
	_numpy_dtypes = numpy.float16, numpy.float32, numpy.float64
	_torch_dtypes = torch.float16, torch.float32, torch.float64

	for _dtype1, _dtype2 in zip(_numpy_dtypes, _torch_dtypes):
		_test_cast(numpy.array([[1.2, 2.0, 3.1], [-1.4, -1.3, -0.0]], 
			dtype=_dtype1), _dtype2, 2)
		_test_cast(numpy.array([[0.0, -3.0, 0.0]], dtype=_dtype1), _dtype2, 2)
		_test_cast(numpy.array([[0.0]], dtype=_dtype1), _dtype2, 2)


def test_cast_as_tensor_numpy_check_wrong_2d():
	x1 = numpy.array([[True, True, False]])
	x2 = numpy.array([[1, 2, 3]])
	x3 = numpy.array([[1.0, 1.1, 1.1]])

	assert_raises(AssertionError, _test_cast, x1, torch.int64, 2)
	assert_raises(AssertionError, _test_cast, x1, torch.bool, 1)
	assert_raises(AssertionError, _test_cast, x2, torch.int32, 2)
	assert_raises(AssertionError, _test_cast, x2, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, x3, torch.int64, 2)
	assert_raises(AssertionError, _test_cast, x3, torch.float64, 1)



def test_cast_as_tensor_torch_bool_1d():
	_test_cast(torch.tensor([True, False, True, True]), torch.bool, 1)
	_test_cast(torch.tensor([True, True, True]), torch.bool, 1)
	_test_cast(torch.tensor([False]), torch.bool, 1)


def test_cast_as_tensor_torch_int_1d():
	_torch_dtypes = torch.int32, torch.int64

	for _dtype in _torch_dtypes:
		_test_cast(torch.tensor([1, 2, 3], dtype=_dtype), _dtype, 1)
		_test_cast(torch.tensor([0, -3, 0], dtype=_dtype), _dtype, 1)
		_test_cast(torch.tensor([0], dtype=_dtype), _dtype, 1)


def test_cast_as_tensor_torch_float_1d():
	_torch_dtypes = torch.float16, torch.float32, torch.float64

	for _dtype in _torch_dtypes:
		_test_cast(torch.tensor([1.2, 2.0, 3.1], dtype=_dtype), _dtype, 1)
		_test_cast(torch.tensor([0.0, -3.0, 0.0], dtype=_dtype), _dtype, 1)
		_test_cast(torch.tensor([0.0], dtype=_dtype), _dtype, 1)


def test_cast_as_tensor_torch_check_wrong_1d():
	x1 = torch.tensor([True, True, False])
	x2 = torch.tensor([1, 2, 3])
	x3 = torch.tensor([1.0, 1.1, 1.1])

	assert_raises(AssertionError, _test_cast, x1, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, x1, torch.bool, 0)
	assert_raises(AssertionError, _test_cast, x2, torch.int32, 1)
	assert_raises(AssertionError, _test_cast, x2, torch.int64, 0)
	assert_raises(AssertionError, _test_cast, x3, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, x3, torch.float64, 0)



def test_cast_as_tensor_torch_bool_2d():
	_test_cast(torch.tensor([[True, False, True], [True, True, False]]), 
		torch.bool, 2)
	_test_cast(torch.tensor([[True, True, True]]), torch.bool, 2)
	_test_cast(torch.tensor([[False]]), torch.bool, 2)


def test_cast_as_tensor_torch_int_2d():
	_torch_dtypes = torch.int32, torch.int64

	for _dtype in _torch_dtypes:
		_test_cast(torch.tensor([[1, 2, 3], [3, 2, 1]], dtype=_dtype), 
			_dtype, 2)
		_test_cast(torch.tensor([[0, -3, 0]], dtype=_dtype), _dtype, 2)
		_test_cast(torch.tensor([[0]], dtype=_dtype), _dtype, 2)


def test_cast_as_tensor_torch_float_2d():
	_torch_dtypes = torch.float16, torch.float32, torch.float64

	for _dtype in _torch_dtypes:
		_test_cast(torch.tensor([[1.2, 2.0, 3.1], [-1.4, -1.3, -0.0]], 
			dtype=_dtype), _dtype, 2)
		_test_cast(torch.tensor([[0.0, -3.0, 0.0]], dtype=_dtype), _dtype, 2)
		_test_cast(torch.tensor([[0.0]], dtype=_dtype), _dtype, 2)


def test_cast_as_tensor_torch_check_wrong_2d():
	x1 = torch.tensor([[True, True, False]])
	x2 = torch.tensor([[1, 2, 3]])
	x3 = torch.tensor([[1.0, 1.1, 1.1]])

	assert_raises(AssertionError, _test_cast, x1, torch.int64, 2)
	assert_raises(AssertionError, _test_cast, x1, torch.bool, 1)
	assert_raises(AssertionError, _test_cast, x2, torch.int32, 2)
	assert_raises(AssertionError, _test_cast, x2, torch.int64, 1)
	assert_raises(AssertionError, _test_cast, x3, torch.int64, 2)
	assert_raises(AssertionError, _test_cast, x3, torch.float64, 1)



#

def _test_update(inertia):
	x1 = torch.tensor([1.0, 1.4, 1.8, -1.1, 0.0])
	x2 = torch.tensor([2.2, 8.2, 0.1, 105.2, 0.0])
	y = x1 * inertia + x2 * (1-inertia) 

	_update_parameter(x1, x2, inertia=inertia)
	assert_array_almost_equal(x1, y)


def test_update_parameter():
	_test_update(inertia=0.0)


def test_update_parameter_inertia():
	_test_update(inertia=0.1)
	_test_update(inertia=0.5)
	_test_update(inertia=1.0)



#



def test_check_parameters_min_values_bool():
	x = torch.tensor([True, True, False], dtype=torch.bool)
	dtypes = [torch.bool]

	_check_parameter(x, "x", min_value=0)
	_check_parameter(x, "x", min_value=-1.0)

	assert_raises(ValueError, _check_parameter, x, "x", min_value=1)
	assert_raises(ValueError, _check_parameter, x, "x", min_value=1000.0)


def test_check_parameters_min_values_int():
	x = torch.tensor([1, 6, 24], dtype=torch.int32)
	dtypes = [torch.bool]

	_check_parameter(x, "x", min_value=1)
	_check_parameter(x, "x", min_value=-1.0)

	assert_raises(ValueError, _check_parameter, x, "x", min_value=2)
	assert_raises(ValueError, _check_parameter, x, "x", min_value=25.0)


def test_check_parameters_min_values_float():
	x = torch.tensor([1, 6, 24], dtype=torch.float32)
	dtypes = [torch.bool]

	_check_parameter(x, "x", min_value=1)
	_check_parameter(x, "x", min_value=-1.0)

	assert_raises(ValueError, _check_parameter, x, "x", min_value=2)
	assert_raises(ValueError, _check_parameter, x, "x", min_value=25.0)


def test_check_parameters_max_values_bool():
	x = torch.tensor([True, True, False], dtype=torch.bool)
	dtypes = [torch.bool]

	_check_parameter(x, "x", max_value=1)
	_check_parameter(x, "x", max_value=100.0)

	assert_raises(ValueError, _check_parameter, x, "x", max_value=0)
	assert_raises(ValueError, _check_parameter, x, "x", max_value=-2.7)


def test_check_parameters_max_values_int():
	x = torch.tensor([1, 6, 24], dtype=torch.int32)
	dtypes = [torch.bool]

	_check_parameter(x, "x", max_value=24)
	_check_parameter(x, "x", max_value=24.1)

	assert_raises(ValueError, _check_parameter, x, "x", max_value=2)
	assert_raises(ValueError, _check_parameter, x, "x", max_value=8.0)


def test_check_parameters_max_values_float():
	x = torch.tensor([1, 6, 24], dtype=torch.float32)
	dtypes = [torch.bool]

	_check_parameter(x, "x", max_value=24)
	_check_parameter(x, "x", max_value=24.1)

	assert_raises(ValueError, _check_parameter, x, "x", max_value=2)
	assert_raises(ValueError, _check_parameter, x, "x", max_value=8.0)


def test_check_parameters_minmax_values_float():
	x = torch.tensor([1.1, 2.3, 7.8], dtype=torch.float32)
	dtypes = [torch.bool]

	_check_parameter(x, "x", min_value=1.0, max_value=24)

	assert_raises(ValueError, _check_parameter, x, "x", min_value=1.2, 
		max_value=24)
	assert_raises(ValueError, _check_parameter, x, "x", min_value=0.0,
		max_value=6)


def test_check_parameters_value_sum_float():
	x = torch.tensor([1.1, 2.3, 7.8], dtype=torch.float32)
	_check_parameter(x, "x", value_sum=torch.sum(x))

	assert_raises(ValueError, _check_parameter, x, "x", value_sum=-0.2)
	assert_raises(ValueError, _check_parameter, x, "x", value_sum=torch.sum(x)
		+2e-6)


	x = 1.5
	_check_parameter(x, "x", value_sum=x)

	assert_raises(ValueError, _check_parameter, x, "x", value_sum=-0.2)
	assert_raises(ValueError, _check_parameter, x, "x", value_sum=x+2e-6)


def test_check_parameters_value_set_bool():
	x = torch.tensor([True, True, True], dtype=torch.bool)
	value_set = [True]

	_check_parameter(x, "x", value_set=tuple(value_set))
	_check_parameter(x, "x", value_set=list(value_set))

	assert_raises(ValueError, _check_parameter, x, "x", value_set=[False])
	assert_raises(ValueError, _check_parameter, x, "x", value_set=[5.2])


def test_check_parameters_value_set_int():
	x = torch.tensor([2, 6, 24], dtype=torch.int32)
	value_set = [2, 6, 24, 26]

	_check_parameter(x, "x", value_set=tuple(value_set))
	_check_parameter(x, "x", value_set=list(value_set))

	assert_raises(ValueError, _check_parameter, x, "x", value_set=[True, False])
	assert_raises(ValueError, _check_parameter, x, "x", value_set=[5.2, 1, 6])


#def test_check_parameters_value_set_float():
#	x = torch.tensor([1.1, 6.0, 24.3], dtype=torch.float32)
#	value_set = [1.1, 6.0, 24.3, 17.8]
#
#	_check_parameter(x, "x", value_set=tuple(value_set))
#	_check_parameter(x, "x", value_set=list(value_set))
#
#	assert_raises(ValueError, _check_parameter, x, "x", value_set=[True, False])
#	assert_raises(ValueError, _check_parameter, x, "x", value_set=[5.2, 1, 6])


def test_check_parameters_dtypes_bool():
	x = torch.tensor([True, True, False], dtype=torch.bool)
	dtypes = [torch.bool]

	_check_parameter(x, "x", dtypes=tuple(dtypes))
	_check_parameter(x, "x", dtypes=list(dtypes))
	_check_parameter(x, "x", dtypes=set(dtypes))

	assert_raises(ValueError, _check_parameter, x, "x", dtypes=[torch.int64])
	assert_raises(ValueError, _check_parameter, x, "x", dtypes=[torch.float64])


def test_check_parameters_dtypes_int():
	x = torch.tensor([1, 2, 3], dtype=torch.int32)
	dtypes = [torch.int32, torch.int64]

	_check_parameter(x, "x", dtypes=tuple(dtypes))
	_check_parameter(x, "x", dtypes=list(dtypes))
	_check_parameter(x, "x", dtypes=set(dtypes))

	assert_raises(ValueError, _check_parameter, x, "x", dtypes=[torch.int64])
	assert_raises(ValueError, _check_parameter, x, "x", dtypes=[torch.float32])


def test_check_parameters_dtypes_float():
	x = torch.tensor([1, 2, 3], dtype=torch.float32)
	dtypes = [torch.float32, torch.float64]

	_check_parameter(x, "x", dtypes=tuple(dtypes))
	_check_parameter(x, "x", dtypes=list(dtypes))
	_check_parameter(x, "x", dtypes=set(dtypes))

	assert_raises(ValueError, _check_parameter, x, "x", dtypes=[torch.int64])
	assert_raises(ValueError, _check_parameter, x, "x", dtypes=[torch.float64])


def test_check_parameters_ndim_0():
	x = torch.tensor(1.1)

	_check_parameter(x, "x", ndim=0)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=2)


def test_check_parameters_ndim_1():
	x = torch.tensor([1.1])

	_check_parameter(x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=0)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=2)


def test_check_parameters_ndim_2():
	x = torch.tensor([[1.1]])

	_check_parameter(x, "x", ndim=2)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=0)


def test_check_parameters_ndim_tuple():
	x = torch.tensor([1.1])

	_check_parameter(x, "x", ndim=(1,))
	_check_parameter(x, "x", ndim=(0, 1))
	_check_parameter(x, "x", ndim=(1, 2))

	assert_raises(ValueError, _check_parameter, x, "x", ndim=0)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=(0, 2))

	x = torch.tensor(1.1)

	_check_parameter(x, "x", ndim=(0,))
	_check_parameter(x, "x", ndim=(0, 1))
	_check_parameter(x, "x", ndim=(0, 2))

	assert_raises(ValueError, _check_parameter, x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=(1, 2))

	x = torch.tensor([[1.1]])

	_check_parameter(x, "x", ndim=(2,))
	_check_parameter(x, "x", ndim=(2, 1))
	_check_parameter(x, "x", ndim=(0, 2))

	assert_raises(ValueError, _check_parameter, x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=(0, 1))


def test_check_parameters_ndim_list():
	x = torch.tensor([1.1])

	_check_parameter(x, "x", ndim=[1])
	_check_parameter(x, "x", ndim=[0, 1])
	_check_parameter(x, "x", ndim=[1, 2])

	assert_raises(ValueError, _check_parameter, x, "x", ndim=0)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=[0, 2])

	x = torch.tensor(1.1)

	_check_parameter(x, "x", ndim=[0])
	_check_parameter(x, "x", ndim=[0, 1])
	_check_parameter(x, "x", ndim=[0, 2])

	assert_raises(ValueError, _check_parameter, x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=[1, 2])

	x = torch.tensor([[1.1]])

	_check_parameter(x, "x", ndim=[2])
	_check_parameter(x, "x", ndim=[2, 1])
	_check_parameter(x, "x", ndim=[0, 2])

	assert_raises(ValueError, _check_parameter, x, "x", ndim=1)
	assert_raises(ValueError, _check_parameter, x, "x", ndim=[0, 1])


def test_check_parameters_shape():
	x = torch.tensor([[1.1]])

	_check_parameter(x, "x", shape=(1, 1))
	_check_parameter(x, "x", shape=(-1, 1))
	_check_parameter(x, "x", shape=(1, -1))
	_check_parameter(x, "x", shape=(-1, -1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(2, 1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1, 2))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1,))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1, 2, 1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1, 2, -1))

	x = torch.tensor([
		[1.1, 1.2, 1.3, 1.8],
		[2.1, 1.1, 1.4, 0.9]
	])

	_check_parameter(x, "x", shape=(2, 4))
	_check_parameter(x, "x", shape=(-1, 4))
	_check_parameter(x, "x", shape=(2, -1))
	_check_parameter(x, "x", shape=(-1, -1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(2, 1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1, 2))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1,))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1, 2, 1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(1, 2, -1))
	assert_raises(ValueError, _check_parameter, x, "x", shape=(2, -1, -1))


###


@pytest.fixture
def X1():
	return [[[0.5, 0.2], 
		     [0.2, 0.1]],

		    [[0.1, 0.2],
		     [0.7, 0.2]],
 
		    [[0.1, 0.4],
		     [0.3, 0.1]],

		    [[0.3, 0.1],
		     [0.1, 0.6]]]


@pytest.fixture
def w1():
	return [[0.1, 0.5], [0.1, 0.3], [0.5, 0.2], [0.2, 0.1]]


@pytest.fixture
def p1():
	return [[[0.5, 0.5], 
		  [0.5, 0.5]],

		 [[0.7, 0.3],
		  [0.3, 0.7]],

		 [[1.0, 0.0],
		  [0.5, 0.5]],

		 [[0.0, 1.0],
		  [0.0, 1.0]]]


def test_partition_3d_X(X1):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		y, _, _ = partition_sequences(func(X1))

		assert isinstance(y, list)
		assert len(y) == 1

		assert isinstance(y[0], torch.Tensor)
		assert y[0].ndim == 3
		assert y[0].shape == (4, 2, 2)
		assert_array_almost_equal(X1, y[0])


def test_partition_3d_Xw(X1, w1):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		y, w, p = partition_sequences(func(X1), sample_weight=func(w1))

		assert isinstance(y, list)
		assert len(y) == 1

		assert isinstance(w, list)
		assert len(w) == 1

		assert p is None

		assert isinstance(y[0], torch.Tensor)
		assert y[0].ndim == 3
		assert y[0].shape == (4, 2, 2)
		assert_array_almost_equal(X1, y[0])

		assert isinstance(w[0], torch.Tensor)
		assert w[0].ndim == 2
		assert w[0].shape == (4, 2)
		assert_array_almost_equal(w1, w[0])


def test_partition_3d_Xp(X1, p1):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		y, w, p = partition_sequences(func(X1), priors=func(p1), n_dists=2)

		assert isinstance(y, list)
		assert len(y) == 1

		assert w is None

		assert isinstance(p, list)
		assert len(p) == 1

		assert isinstance(y[0], torch.Tensor)
		assert y[0].ndim == 3
		assert y[0].shape == (4, 2, 2)
		assert_array_almost_equal(X1, y[0])

		assert isinstance(p[0], torch.Tensor)
		assert p[0].ndim == 3
		assert p[0].shape == (4, 2, 2)
		assert_array_almost_equal(p1, p[0])


def test_partition_3d_Xwp(X1, w1, p1):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		y, w, p = partition_sequences(func(X1), sample_weight=func(w1),
									  priors=func(p1), n_dists=2)

		assert isinstance(y, list)
		assert len(y) == 1

		assert isinstance(w, list)
		assert len(w) == 1

		assert isinstance(p, list)
		assert len(p) == 1

		assert isinstance(y[0], torch.Tensor)
		assert y[0].ndim == 3
		assert y[0].shape == (4, 2, 2)
		assert_array_almost_equal(X1, y[0])

		assert isinstance(w[0], torch.Tensor)
		assert w[0].ndim == 2
		assert w[0].shape == (4, 2)
		assert_array_almost_equal(w1, w[0])

		assert isinstance(p[0], torch.Tensor)
		assert p[0].ndim == 3
		assert p[0].shape == (4, 2, 2)
		assert_array_almost_equal(p1, p[0])


@pytest.fixture
def X2():
	return [[[[0.5, 0.2], 
		      [0.2, 0.1]],

		     [[0.1, 0.2],
		      [0.7, 0.2]]],

		    [[[0.1, 0.4],
		      [0.3, 0.1],
		      [0.1, 0.7]],

		     [[0.3, 0.1],
		      [0.1, 0.6],
		      [0.1, 0.1]]]]


@pytest.fixture
def w2():
	return [[[0.1, 0.5], [0.1, 0.3]], [[0.5, 0.2, 0.1], [0.2, 0.1, 0.0]]]


@pytest.fixture
def p2():
	return [[[[0.5, 0.5], 
		      [1.0, 0.0]],

		     [[0.5, 0.5],
		      [0.5, 0.5]]],

		    [[[0.5, 0.5],
		      [0.7, 0.3],
		      [0.1, 0.9]],

		     [[1.0, 0.0],
		      [1.0, 0.0],
		      [1.0, 0.0]]]]


def test_partition_3ds_X(X2):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X2_ = [func(x) for x in X2]

		y, _, _ = partition_sequences(X2)

		assert isinstance(y, list)
		assert len(y) == 2

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == (2, i+2, 2)
			assert_array_almost_equal(y_, X2[i])


def test_partition_3ds_Xw(X2, w2):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X2_ = [func(x) for x in X2]
		w2_ = [func(w) for w in w2]

		y, w, p = partition_sequences(X2_, sample_weight=w2_) 

		assert isinstance(y, list)
		assert len(y) == 2

		assert isinstance(w, list)
		assert len(w) == 2

		assert p == None

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == (2, i+2, 2)
			assert_array_almost_equal(y_, X2[i])

		for i, w_ in enumerate(w):
			assert isinstance(w_, torch.Tensor)
			assert w_.ndim == 2
			assert w_.shape == (2, i+2)
			assert_array_almost_equal(w_, w2[i])


def test_partition_3ds_Xp(X2, p2):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X2_ = [func(x) for x in X2]
		p2_ = [func(p) for p in p2]

		y, w, p = partition_sequences(X2_, priors=p2_, n_dists=2)

		assert isinstance(y, list)
		assert len(y) == 2

		assert isinstance(p, list)
		assert len(p) == 2

		assert w is None

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == (2, i+2, 2)
			assert_array_almost_equal(y_, X2[i])

		for i, p_ in enumerate(p):
			assert isinstance(p_, torch.Tensor)
			assert p_.ndim == 3
			assert p_.shape == (2, i+2, 2)
			assert_array_almost_equal(p_, p2[i])


def test_partition_3ds_Xwp(X2, w2, p2):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X2_ = [func(x) for x in X2]
		w2_ = [func(w) for w in w2]
		p2_ = [func(p) for p in p2]

		y, w, p = partition_sequences(X2_,
			sample_weight=w2_, priors=p2_, n_dists=2)

		assert isinstance(y, list)
		assert len(y) == 2

		assert isinstance(p, list)
		assert len(p) == 2

		assert isinstance(w, list)
		assert len(w) == 2

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == (2, i+2, 2)
			assert_array_almost_equal(y_, X2[i])

		for i, w_ in enumerate(w):
			assert isinstance(w_, torch.Tensor)
			assert w_.ndim == 2
			assert w_.shape == (2, i+2)
			assert_array_almost_equal(w_, w2[i])

		for i, p_ in enumerate(p):
			assert isinstance(p_, torch.Tensor)
			assert p_.ndim == 3
			assert p_.shape == (2, i+2, 2)
			assert_array_almost_equal(p_, p2[i])


@pytest.fixture
def X3():
	return [[[0.5, 0.2], 
		     [0.2, 0.1]],

		    [[0.1, 0.2],
		     [0.7, 0.2]],

		    [[0.1, 0.4],
		     [0.3, 0.1],
		     [0.1, 0.7]],

		    [[0.3, 0.1],
		     [0.1, 0.6],
		     [0.1, 0.1]],

		    [[0.5, 0.1]],

		    [[0.5, 0.2],
		     [0.3, 0.4]]]


@pytest.fixture
def w3():
	return [[0.1, 0.5], [0.1, 0.3], [0.5, 0.2, 0.1], [0.2, 0.1, 0.3],
		[0.1], [0.6, 0.4]]


@pytest.fixture
def p3():
	return [[[0.5, 0.5], 
		     [0.5, 0.5]],

		    [[0.9, 0.1],
		     [0.0, 1.0]],

		    [[0.5, 0.5],
		     [0.5, 0.5],
		     [0.0, 1.0]],

		    [[0.0, 1.0],
		     [0.0, 1.0],
		     [1.0, 0.0]],

		    [[0.5, 0.5]],

		    [[0.7, 0.3],
		     [0.3, 0.7]]]


def test_partition_2ds_X(X3):
	X_batches = [
		[[[0.5, 0.1]]],

		[[[0.5, 0.2], 
		  [0.2, 0.1]],

		 [[0.1, 0.2],
		  [0.7, 0.2]],

		 [[0.5, 0.2],
		  [0.3, 0.4]]],

		[[[0.1, 0.4],
		  [0.3, 0.1],
		  [0.1, 0.7]],

		  [[0.3, 0.1],
		   [0.1, 0.6],
		   [0.1, 0.1]]]]


	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X3_ = [func(x) for x in X3]

		y, _, _ = partition_sequences(X3_)

		assert isinstance(y, list)
		assert len(y) == 3

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == ([1, 3, 2][i], i+1, 2)
			assert_array_almost_equal(y_, X_batches[i])


def test_partition_2ds_Xw(X3, w3):
	w_batches = [[[0.1]], [[0.1, 0.5], [0.1, 0.3], [0.6, 0.4]], 
		[[0.5, 0.2, 0.1], [0.2, 0.1, 0.3]]]


	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X3_ = [func(x) for x in X3]
		w3_ = [func(w) for w in w3]

		y, w, p = partition_sequences(X3_, sample_weight=w3_) 

		assert isinstance(y, list)
		assert len(y) == 3

		assert isinstance(w, list)
		assert len(w) == 3

		assert p == None

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == ([1, 3, 2][i], i+1, 2)

		for i, w_ in enumerate(w):
			assert isinstance(w_, torch.Tensor)
			assert w_.ndim == 2
			assert w_.shape == ([1, 3, 2][i], i+1)
			assert_array_almost_equal(w_, w_batches[i])


def test_partition_2ds_Xp(X3, p3):
	p_batches = [
		[[[0.5, 0.5]]],

		[[[0.5, 0.5], 
		  [0.5, 0.5]],

		 [[0.9, 0.1],
		  [0.0, 1.0]],

		 [[0.7, 0.3],
		  [0.3, 0.7]]],

		[[[0.5, 0.5],
		  [0.5, 0.5],
		  [0.0, 1.0]],

		 [[0.0, 1.0],
		  [0.0, 1.0],
		  [1.0, 0.0]]]]

	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X3_ = [func(x) for x in X3]
		p3_ = [func(p) for p in p3]

		y, w, p = partition_sequences(X3_, priors=p3_, n_dists=2)

		assert isinstance(y, list)
		assert len(y) == 3

		assert isinstance(p, list)
		assert len(p) == 3

		assert w is None

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == ([1, 3, 2][i], i+1, 2)

		for i, p_ in enumerate(p):
			assert isinstance(p_, torch.Tensor)
			assert p_.ndim == 3
			assert p_.shape == ([1, 3, 2][i], i+1, 2)
			assert_array_almost_equal(p_, p_batches[i])


def test_partition_2ds_Xwp(X3, w3, p3):
	funcs = (lambda x: x, tuple, numpy.array, 
		lambda x: torch.from_numpy(numpy.array(x)))

	for func in funcs:
		X3_ = [func(x) for x in X3]
		w3_ = [func(w) for w in w3]
		p3_ = [func(p) for p in p3]

		y, w, p = partition_sequences(X3_, sample_weight=w3_, priors=p3_, n_dists=2)

		assert isinstance(y, list)
		assert len(y) == 3

		assert isinstance(p, list)
		assert len(w) == 3

		assert isinstance(p, list)
		assert len(p) == 3

		for i, y_ in enumerate(y):
			assert isinstance(y_, torch.Tensor)
			assert y_.ndim == 3
			assert y_.shape == ([1, 3, 2][i], i+1, 2)

		for i, w_ in enumerate(w):
			assert isinstance(w_, torch.Tensor)
			assert w_.ndim == 2
			assert w_.shape == ([1, 3, 2][i], i+1)

		for i, p_ in enumerate(p):
			assert isinstance(p_, torch.Tensor)
			assert p_.ndim == 3
			assert p_.shape == ([1, 3, 2][i], i+1, 2)


@pytest.mark.parametrize("X", [torch.ones((1, 10, 2)), [torch.ones((1, 10, 2)), torch.ones((2, 10, 2))]])
@pytest.mark.parametrize("invalid", ["sample_weight", "priors"])
def test_dont_hide_errors_for_priors_and_sample_weight(X, invalid):
	"""Test that we get the correct error message when we don't pass data in case 3."""

	with pytest.raises(ValueError) as excinfo:
		partition_sequences(X, **{invalid: numpy.zeros((1, 1)) - 1}, n_dists=10)

	assert invalid in str(excinfo.value)


from .bernoulli import Bernoulli
from .categorical import Categorical
from .conditional_categorical import ConditionalCategorical
from .dirac_delta import DiracDelta
from .exponential import Exponential
from .gamma import Gamma
from .independent_components import IndependentComponents
from .joint_categorical import JointCategorical
from .normal import Normal
from .poisson import Poisson
from .student_t import StudentT
from .uniform import Uniform
from .zero_inflated import ZeroInflated
from .lognormal import LogNormal
from .halfnormal import HalfNormal


# kmeans.py
# Author: Jacob Schreiber

import time
import torch

from ._utils import _cast_as_tensor
from ._utils import _cast_as_parameter
from ._utils import _update_parameter
from ._utils import _check_parameter
from ._utils import _reshape_weights
from ._utils import _initialize_centroids

from ._utils import eps

class KMeans(torch.nn.Module):
	"""A KMeans clustering object.

	Although K-means clustering is not a probabilistic model by itself,
	necessarily, it can be a useful initialization for many probabilistic
	methods, and can also be thought of as a specific formulation of a mixture
	model. Specifically, if you have a Gaussian mixture model with diagonal
	covariances set to 1/inf, in theory you will get the exact same results
	as k-means clustering.

	The implementation is provided here not necessarily to compete with other
	implementations, but simply to use as a consistent initialization.


	Parameters
	----------
	k: int or None, optional
		The number of clusters to initialize. Default is None

	centroids: list, numpy.ndarray, torch.Tensor, or None, optional
		A set of centroids to use to initialize the algorithm. Default is None.

	init: str, optional
		The initialization to use if `centroids` are not provided. Default is
		'first-k'. Must be one of:

			'first-k': Use the first k examples from the data set
			'random': Use a random set of k examples from the data set
			'submodular-facility-location': Use a facility location submodular
				objective to initialize the k-means algorithm
			'submodular-feature-based': Use a feature-based submodular objective
				to initialize the k-means algorithm.

	max_iter: int, optional
		The number of iterations to do in the EM step of fitting the
		distribution. Default is 10.

	tol: float, optional
		The threshold at which to stop during fitting when the improvement
		goes under. Default is 0.1.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	verbose: bool, optional
		Whether to print the improvement and timings during training.
	"""

	def __init__(self, k=None, centroids=None, init='first-k', max_iter=10, 
		tol=0.1, inertia=0.0, frozen=False, random_state=None, verbose=False):
		super().__init__()
		self.name = "KMeans"
		self._device = _cast_as_parameter([0.0])

		self.centroids = _check_parameter(_cast_as_parameter(centroids, 
			dtype=torch.float32), "centroids", ndim=2)

		self.k = _check_parameter(_cast_as_parameter(k), "k", ndim=0, 
			min_value=2, dtypes=(int, torch.int32, torch.int64))

		self.init = _check_parameter(init, "init", value_set=("random", 
			"first-k", "submodular-facility-location", 
			"submodular-feature-based"), ndim=0, dtypes=(str,))
		self.max_iter = _check_parameter(_cast_as_tensor(max_iter), "max_iter",
			ndim=0, min_value=1, dtypes=(int, torch.int32, torch.int64))
		self.tol = _check_parameter(_cast_as_tensor(tol), "tol", ndim=0,
			min_value=0)
		self.inertia = _check_parameter(_cast_as_tensor(inertia), "inertia",
			ndim=0, min_value=0.0, max_value=1.0)
		self.frozen = _check_parameter(_cast_as_tensor(frozen), "frozen",
			ndim=0, value_set=(True, False))
		self.random_state = random_state
		self.verbose = _check_parameter(verbose, "verbose", 
			value_set=(True, False))

		if self.k is None and self.centroids is None:
			raise ValueError("Must specify one of `k` or `centroids`.")

		self.k = len(centroids) if centroids is not None else self.k
		self.d = len(centroids[0]) if centroids is not None else None
		self._initialized = centroids is not None
		self._reset_cache()

	@property
	def device(self):
		try:
			return next(self.parameters()).device
		except:
			return 'cpu'

	def _initialize(self, X):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			The data to use to initialize the model.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2)
		centroids = _initialize_centroids(X, self.k, algorithm=self.init, 
			random_state=self.random_state)
		
		if isinstance(centroids, torch.masked.MaskedTensor):
			centroids = centroids._masked_data * centroids._masked_mask

		if centroids.device != self.device:
			centroids = centroids.to(self.device)

		self.centroids = _cast_as_parameter(centroids)

		self.d = X.shape[1]
		self._initialized = True
		self._reset_cache()

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		self.register_buffer("_w_sum", torch.zeros(self.k, self.d, 
			device=self.device))
		self.register_buffer("_xw_sum", torch.zeros(self.k, self.d,
			device=self.device))
		self.register_buffer("_centroid_sum", torch.sum(self.centroids**2, 
			dim=1).unsqueeze(0))

	def _distances(self, X):
		"""Calculate the distances between each example and each centroid.

		This method calculates the distance between each example and each
		centroid in the model. These distances make up the backbone of the
		k-means learning and prediction steps.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.


		Returns
		-------
		distances: torch.Tensor, shape=(-1, self.k)
			The Euclidean distance between each example and each cluster.
		"""

		X = _check_parameter(_cast_as_tensor(X, dtype=torch.float32), "X", 
			ndim=2, shape=(-1, self.d))

		XX = torch.sum(X**2, dim=1).unsqueeze(1)

		if isinstance(X, torch.masked.MaskedTensor):
			n = X._masked_mask.sum(dim=1).unsqueeze(1)
			Xc = torch.matmul(X._masked_data * X._masked_mask, self.centroids.T)
		else:
			n = X.shape[1]
			Xc = torch.matmul(X, self.centroids.T)
		
		distances = torch.empty(X.shape[0], self.k, dtype=X.dtype, 
			device=self.device)
		distances[:] = torch.clamp(XX - 2*Xc + self._centroid_sum, min=0)
		return torch.sqrt(distances / n)

	def predict(self, X):
		"""Calculate the cluster assignment for each example.

		This method calculates cluster assignment for each example as the
		nearest centroid according to the Euclidean distance.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.


		Returns
		-------
		y: torch.Tensor, shape=(-1,)
			The predicted label for each example.
		"""

		return self._distances(X).argmin(dim=1)

	def summarize(self, X, sample_weight=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.

		For k-means clustering, this step is essentially performing the 'E' part
		of the EM algorithm on a batch of data, where examples are hard-assigned
		to distributions in the model and summaries are derived from that.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.
		"""

		if self.frozen:
			return 0

		if not self._initialized:
			self._initialize(X)

		X = _check_parameter(_cast_as_tensor(X, dtype=torch.float32), "X", 
			ndim=2, shape=(-1, self.d))
		sample_weight = _reshape_weights(X, _cast_as_tensor(sample_weight, 
			dtype=torch.float32), device=self.device)

		distances = self._distances(X)
		y_hat = distances.argmin(dim=1)

		if isinstance(X, torch.masked.MaskedTensor):
			for i in range(self.k):
				idx = y_hat == i

				self._w_sum[i][:] = self._w_sum[i] + sample_weight[idx].sum(dim=0)
				self._xw_sum[i][:] = self._xw_sum[i] + (X[idx] * 
					sample_weight[idx]).sum(dim=0)

		else:
			y_hat = y_hat.unsqueeze(1).expand(-1, self.d)
			self._w_sum.scatter_add_(0, y_hat, sample_weight)
			self._xw_sum.scatter_add_(0, y_hat, X * sample_weight)
		
		return distances.min(dim=1).values.sum()

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		if self.frozen:
			return

		centroids = self._xw_sum / self._w_sum
		_update_parameter(self.centroids, centroids, self.inertia)
		self._reset_cache()

	def fit(self, X, sample_weight=None):
		"""Fit the model to optionally weighted examples.

		This method implements the core of the learning process. For a
		mixture model, this involves performing EM until the distributions that
		are being fit converge according to the threshold set by `tol`, or
		until the maximum number of iterations has been hit.

		This method is largely a wrapper around the `summarize` and
		`from_summaries` methods. It's primary contribution is serving as a
		loop around these functions and to monitor convergence.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		self
		"""

		d_current = None
		for i in range(self.max_iter):
			start_time = time.time()

			d_previous = d_current
			d_current = self.summarize(X, sample_weight=sample_weight)

			if i > 0:
				improvement = d_previous - d_current
				duration = time.time() - start_time

				if self.verbose:
					print("[{}] Improvement: {}, Time: {:4.4}s".format(i, 
						improvement, duration))

				if improvement < self.tol:
					break

			self.from_summaries()

		self._reset_cache()
		return self

	def fit_predict(self, X, sample_weight=None):
		"""Fit the model and then return the predictions.

		This function wraps a call to the `fit` function and then to the
		`predict` function. Essentially, k-means will be fit to the data and
		the resulting clustering assignments will be returned.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate. 

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.


		Returns
		-------
		y: torch.Tensor, shape=(-1,)
			Cluster assignments for each example.
		"""

		self.fit(X, sample_weight=sample_weight)
		return self.predict(X)


# student_t.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import math
import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter

from .normal import Normal


class StudentT(Normal):
	"""A Student T distribution.

	A Student T distribution models the probability of a variable occurring under
	a bell-shaped curve with heavy tails. Basically, this is a version of the
	normal distribution that is less resistant to outliers.  It is described by 
	a vector of mean values and a vector of variance values. This
	distribution can assume that features are independent of the others if
	the covariance type is 'diag' or 'sphere', but if the type is 'full' then
	the features are not independent.

	There are two ways to initialize this object. The first is to pass in
	the tensor of probability parameters, at which point they can immediately be
	used. The second is to not pass in the rate parameters and then call
	either `fit` or `summary` + `from_summaries`, at which point the probability
	parameter will be learned from data.


	Parameters
	----------
	means: list, numpy.ndarray, torch.Tensor or None, shape=(d,), optional
		The mean values of the distributions. Default is None.

	covs: list, numpy.ndarray, torch.Tensor, or None, optional
		The variances and covariances of the distribution. If covariance_type
		is 'full', the shape should be (self.d, self.d); if 'diag', the shape
		should be (self.d,); if 'sphere', it should be (1,). Note that this is
		the variances or covariances in all settings, and not the standard
		deviation, as may be more common for diagonal covariance matrices.
		Default is None.

	covariance_type: str, optional
		The type of covariance matrix. Must be one of 'full', 'diag', or
		'sphere'. Default is 'full'. 

	min_cov: float or None, optional
		The minimum variance or covariance.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.
	"""

	def __init__(self, dofs, means=None, covs=None, covariance_type='diag', 
		min_cov=None, inertia=0.0, frozen=False, check_data=True):
		dofs = _check_parameter(_cast_as_tensor(dofs), "dofs", min_value=1,
			ndim=0, dtypes=(torch.int32, torch.int64))
		self.dofs = dofs

		super().__init__(means=means, covs=covs, min_cov=min_cov,
			covariance_type=covariance_type, inertia=inertia, frozen=frozen,
			check_data=check_data)

		self.name = "StudentT"

		del self.dofs

		self.register_buffer("dofs", _cast_as_tensor(dofs))
		self.register_buffer("_lgamma_dofsp1", torch.lgamma((dofs + 1) / 2.0))
		self.register_buffer("_lgamma_dofs", torch.lgamma(dofs / 2.0))

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		super()._reset_cache()
		if self._initialized == False:
			return

		self.register_buffer("_log_sqrt_dofs_pi_cov", torch.log(torch.sqrt(
			self.dofs * math.pi * self.covs)))

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: torch.tensor, shape=(n, self.d)
			Randomly generated samples.
		"""

		return torch.distributions.StudentT(self.means, self.covs).sample([n])

	def log_probability(self, X):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 2D
		format.

		Note: This differs from some other log probability calculation
		functions, like those in torch.distributions, because it is not
		returning the log probability of each feature independently, but rather
		the total log probability of the entire example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=2, 
			shape=(-1, self.d), check_parameter=self.check_data)

		t = (X - self.means) ** 2 / self.covs
		return torch.sum(self._lgamma_dofsp1 - self._lgamma_dofs - \
			self._log_sqrt_dofs_pi_cov -((self.dofs + 1) / 2.0) * 
			torch.log(1 + t / self.dofs), dim=-1)




# normal.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter
from .._utils import _check_shapes

from .normal import Normal


class LogNormal(Normal):
    """A lognormal object.

    The parameters are the mu and sigma of the normal distribution, which
    is the the exponential of the log normal distribution. This
    distribution can assume that features are independent of the others if
    the covariance type is 'diag' or 'sphere', but if the type is 'full' then
    the features are not independent.

    There are two ways to initialize this object. The first is to pass in
    the tensor of probablity parameters, at which point they can immediately be
    used. The second is to not pass in the rate parameters and then call
    either `fit` or `summarize` + `from_summaries`, at which point the probability
    parameter will be learned from data.


    Parameters
    ----------
    means: list, numpy.ndarray, torch.Tensor or None, shape=(d,), optional
            The mean values of the normal distributions. Default is None.

    covs: list, numpy.ndarray, torch.Tensor, or None, optional
            The variances and covariances of the distribution. If covariance_type
            is 'full', the shape should be (self.d, self.d); if 'diag', the shape
            should be (self.d,); if 'sphere', it should be (1,). Note that this is
            the variances or covariances in all settings, and not the standard
            deviation, as may be more common for diagonal covariance matrices.
            Default is None.

    covariance_type: str, optional
            The type of covariance matrix. Must be one of 'full', 'diag', or
            'sphere'. Default is 'full'.

    min_cov: float or None, optional
            The minimum variance or covariance.

    inertia: float, [0, 1], optional
            Indicates the proportion of the update to apply to the parameters
            during training. When the inertia is 0.0, the update is applied in
            its entirety and the previous parameters are ignored. When the
            inertia is 1.0, the update is entirely ignored and the previous
            parameters are kept, equivalently to if the parameters were frozen.

    frozen: bool, optional
            Whether all the parameters associated with this distribution are frozen.
            If you want to freeze individual pameters, or individual values in those
            parameters, you must modify the `frozen` attribute of the tensor or
            parameter directly. Default is False.
    """

    def __init__(
        self,
        means=None,
        covs=None,
        covariance_type="full",
        min_cov=None,
        inertia=0.0,
        frozen=False,
        check_data=True,
    ):
        self.name = "LogNormal"
        super().__init__(
            means=means,
            covs=covs,
            covariance_type=covariance_type,
            min_cov=min_cov,
            inertia=inertia,
            frozen=frozen,
            check_data=check_data,
        )

    def sample(self, n):
        """Sample from the probability distribution.

        This method will return `n` samples generated from the underlying
        probability distribution.


        Parameters
        ----------
        n: int
                The number of samples to generate.


        Returns
        -------
        X: torch.tensor, shape=(n, self.d)
                Randomly generated samples.
        """

        if self.covariance_type == "diag":
            return torch.distributions.Normal(self.means, self.covs).sample([n]).exp()
        elif self.covariance_type == "full":
            return (
                torch.distributions.MultivariateNormal(self.means, self.covs)
                .sample([n])
                .exp()
            )

    def log_probability(self, X):
        """Calculate the log probability of each example.

        This method calculates the log probability of each example given the
        parameters of the distribution. The examples must be given in a 2D
        format.

        Note: This differs from some other log probability calculation
        functions, like those in torch.distributions, because it is not
        returning the log probability of each feature independently, but rather
        the total log probability of the entire example.


        Parameters
        ----------
        X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
                A set of examples to evaluate.


        Returns
        -------
        logp: torch.Tensor, shape=(-1,)
                The log probability of each example.
        """

        X = _check_parameter(
            _cast_as_tensor(X, dtype=self.means.dtype),
            "X",
            ndim=2,
            shape=(-1, self.d),
            check_parameter=self.check_data,
        )

        # take the log of X
        x_log = X.log()

        return super().log_probability(x_log)

    def summarize(self, X, sample_weight=None):
        """Extract the sufficient statistics from a batch of data.

        This method calculates the sufficient statistics from optionally
        weighted data and adds them to the stored cache. The examples must be
        given in a 2D format. Sample weights can either be provided as one
        value per example or as a 2D matrix of weights for each feature in
        each example.


        Parameters
        ----------
        X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
                A set of examples to summarize.

        sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
                A set of weights for the examples. This can be either of shape
                (-1, self.d) or a vector of shape (-1,). Default is ones.
        """

        if self.frozen is True:
            return
        X = _cast_as_tensor(X, dtype=self.means.dtype)
        super().summarize(X.log(), sample_weight=sample_weight)



# _base.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import time
import numpy
import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _check_parameter
from .._utils import partition_sequences

from ..distributions._distribution import Distribution

from ..kmeans import KMeans


NEGINF = float("-inf")


def _check_inputs(model, X, emissions, priors):
	if X is None and emissions is None:
		raise ValueError("Must pass in one of `X` or `emissions`.")

	emissions = _check_parameter(_cast_as_tensor(emissions), "emissions", 
		ndim=3)
	if emissions is None:
		emissions = model._emission_matrix(X, priors=priors)

	return emissions


class Silent(Distribution):
	def __init__(self):
		super().__init__(inertia=0.0, frozen=False, check_data=True)


class _BaseHMM(Distribution):
	"""A base hidden Markov model.

	A hidden Markov model is an extension of a mixture model to sequences by
	including a transition matrix between the elements of the mixture. Each of
	the algorithms for a hidden Markov model are essentially just a revision
	of those algorithms to incorporate this transition matrix.

	There are two main ways one can implement a hidden Markov model: either the
	transition matrix can be implemented in a dense, or a sparse, manner. If the
	transition matrix is dense, implementing it in a dense manner allows for
	the primary computation to use matrix multiplications which can be very
	fast. However, if the matrix is sparse, these matrix multiplications will
	be fairly slow and end up significantly slower than the sparse version of
	a matrix multiplication.

	This object is a wrapper for both implementations, which can be specified
	using the `kind` parameter. Choosing the right implementation will not
	effect the accuracy of the results but will change the speed at which they
	are calculated. 	

	Separately, there are two ways to instantiate the hidden Markov model. The
	first is by passing in a set of distributions, a dense transition matrix, 
	and optionally start/end probabilities. The second is to initialize the
	object without these and then to add edges using the `add_edge` method
	and to add nodes using the `add_nodes` method. Importantly, the way that
	you choose to initialize the hidden Markov model is independent of the
	implementation that you end up choosing. If you pass in a dense transition
	matrix, this will be converted to a sparse matrix with all the zeros
	dropped if you choose `kind='sparse'`.


	Parameters
	----------
	distributions: tuple or list
		A set of distribution objects. These objects do not need to be
		initialized, i.e., can be "Normal()". 

	edges: numpy.ndarray, torch.Tensor, or None. shape=(k,k), optional
		A dense transition matrix of probabilities for how each node or
		distribution passed in connects to each other one. This can contain
		many zeroes, and when paired with `kind='sparse'`, will drop those
		elements from the matrix. Default is None.

	starts: list, numpy.ndarray, torch.Tensor, or None. shape=(k,), optional
		The probability of starting at each node. If not provided, assumes
		these probabilities are uniform. Default is None.

	ends: list, numpy.ndarray, torch.Tensor, or None. shape=(k,), optional
		The probability of ending at each node. If not provided, assumes
		these probabilities are uniform. Default is None.

	kind: str, 'sparse' or 'dense', optional
		The underlying implementation of the transition matrix to use.
		Default is 'sparse'. 

	init: str, optional
		The initialization to use for the k-means initialization approach.
		Default is 'first-k'. Must be one of:

			'first-k': Use the first k examples from the data set
			'random': Use a random set of k examples from the data set
			'submodular-facility-location': Use a facility location submodular
				objective to initialize the k-means algorithm
			'submodular-feature-based': Use a feature-based submodular objective
				to initialize the k-means algorithm.

	max_iter: int, optional
		The number of iterations to do in the EM step, which for HMMs is
		sometimes called Baum-Welch. Default is 10.

	tol: float, optional
		The threshold at which to stop during fitting when the improvement
		goes under. Default is 0.1.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	random_state: int or None, optional
		The random state to make randomness deterministic. If None, not
		deterministic. Default is None.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling.

	verbose: bool, optional
		Whether to print the improvement and timings during training.
	"""

	def __init__(self, distributions=None, starts=None, ends=None, 
		init='random', max_iter=1000, tol=0.1, sample_length=None, 
		return_sample_paths=False, inertia=0.0, frozen=False, check_data=True,
		random_state=None, verbose=False):
		super().__init__(inertia=inertia, frozen=frozen, check_data=check_data)

		self.distributions = distributions
		n = len(distributions) if distributions is not None else None

		self.start = Silent()
		self.end = Silent()

		self.edges = None
		self.starts = None
		self.ends = None

		if starts is not None:
			starts = _check_parameter(_cast_as_tensor(starts), "starts", ndim=1, 
				shape=(n,), min_value=0., max_value=1., value_sum=1.0)
			self.starts = _cast_as_parameter(torch.log(starts))

		if ends is not None:
			ends = _check_parameter(_cast_as_tensor(ends), "ends", ndim=1, 
				shape=(n,), min_value=0., max_value=1.)
			self.ends = _cast_as_parameter(torch.log(ends))

		if not isinstance(random_state, numpy.random.RandomState):
			self.random_state = numpy.random.RandomState(random_state)
		else:
			self.random_state = random_state

		self.init = init
		self.max_iter = _check_parameter(max_iter, "max_iter", min_value=1, 
			ndim=0, dtypes=(int, torch.int32, torch.int64))
		self.tol = _check_parameter(tol, "tol", min_value=0., ndim=0)

		self.sample_length = sample_length
		self.return_sample_paths = return_sample_paths

		self.verbose = verbose
		self.d = self.distributions[0].d if distributions is not None else None

	def _initialize(self, X=None, sample_weight=None):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d), optional
			The data to use to initialize the model. Default is None.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, len) or a vector of shape (-1,). If None, defaults to ones.
			Default is None.
		"""

		n = self.n_distributions
		if self.starts is None:
			self.starts = _cast_as_parameter(torch.log(torch.ones(n, 
				dtype=self.dtype, device=self.device) / n))

		if self.ends is None:
			self.ends = _cast_as_parameter(torch.log(torch.ones(n,
				dtype=self.dtype, device=self.device) / n))

		_init = all(d._initialized for d in self.distributions)
		if X is not None and not _init:
			if isinstance(X, list):
				d = _cast_as_tensor(X[0]).shape[-1]
				X = torch.cat([_cast_as_tensor(x).reshape(-1, d) for x in X],
					dim=0).unsqueeze(0)

			X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
				check_parameter=self.check_data)
			X = X.reshape(-1, X.shape[-1])

			if sample_weight is None:
				sample_weight = torch.ones(1, dtype=self.dtype, 
					device=self.device).to(X.device).expand(X.shape[0], 1)
			else:
				if isinstance(sample_weight, list):
					sample_weight = torch.cat([_cast_as_tensor(w).reshape(-1, 1) 
						for w in sample_weight], dim=0).to(X.device)

				sample_weight = _check_parameter(_cast_as_tensor(
					sample_weight).reshape(-1, 1), "sample_weight", 
					min_value=0., ndim=1, shape=(len(X),), 
					check_parameter=self.check_data).reshape(-1, 1)

			model = KMeans(self.n_distributions, init=self.init, max_iter=1, 
				random_state=self.random_state).to(self.device)

			y_hat = model.fit_predict(X, sample_weight=sample_weight)

			for i in range(self.n_distributions):
				self.distributions[i].fit(X[y_hat == i].cpu(), 
					sample_weight=sample_weight[y_hat == i].cpu())
				self.distributions[i].to(X.device)

			self.d = X.shape[-1]
			super()._initialize(X.shape[-1])

		self._initialized = True
		self._reset_cache()

	def _emission_matrix(self, X, priors=None):
		"""Return the emission/responsibility matrix.

		This method returns the log probability of each example under each
		distribution contained in the model with the log prior probability
		of each component added.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to evaluate. 

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.

	
		Returns
		-------
		e: torch.Tensor, shape=(-1, len, self.k)
			A set of log probabilities for each example under each distribution.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			shape=(-1, -1, self.d), check_parameter=self.check_data)

		n, k, _ = X.shape
		X = X.reshape(n*k, self.d)

		priors = _check_parameter(_cast_as_tensor(priors), "priors",
			ndim=3, shape=(n, k, self.k), min_value=0.0, max_value=1.0, 
			value_sum=1.0, value_sum_dim=-1, check_parameter=self.check_data)

		if not self._initialized:
			self._initialize()

		e = torch.empty((k, self.k, n), dtype=self.dtype, requires_grad=False, 
			device=self.device)
		
		for i, node in enumerate(self.distributions):
			logp = node.log_probability(X)
			if isinstance(logp, torch.masked.MaskedTensor):
				logp = logp._masked_data

			e[:, i] = logp.reshape(n, k).T

		e = e.permute(2, 0, 1)
		if priors is not None:
			e += torch.log(priors)

		return e

	@property
	def k(self):
		return len(self.distributions) if self.distributions is not None else 0

	@property
	def n_distributions(self):
		return len(self.distributions) if self.distributions is not None else 0

	def freeze(self):
		"""Freeze this model and all child distributions."""

		self.register_buffer("frozen", _cast_as_tensor(True))
		for d in self.distributions:
			d.freeze()

		return self

	def unfreeze(self):
		"""Unfreeze this model and all child distributions."""

		self.register_buffer("frozen", _cast_as_tensor(False))
		for d in self.distributions:
			d.unfreeze()

		return self

	def add_distribution(self, distribution):
		"""Add a distribution to the model.


		Parameters
		----------
		distribution: torchegranate.distributions.Distribution
			A distribution object.
		"""

		if self.distributions is None:
			self.distributions = []

		if not isinstance(distribution, Distribution):
			raise ValueError("distribution must be a distribution object.")

		self.distributions.append(distribution)
		self.d = distribution.d

	def add_distributions(self, distributions):
		"""Add a set of distributions to the model.

		This method will iterative call the `add_distribution`.


		Parameters
		----------
		distributions: list, tuple, iterable
			A set of distributions to add to the model.
		"""

		for distribution in distributions:
			self.add_distribution(distribution)

	def add_edge(self, start, end, probability):
		"""Add an edge to the model.

		This method takes in two distribution objects and the probability
		connecting the two and adds an edge to the model.


		Parameters
		----------
		start: torchegranate.distributions.Distribution
			The parent node for the edge

		end: torchegranate.distributions.Distribution
			The child node for the edge

		probability: float, (0, 1]
			The probability of connecting the two.
		"""

		if not isinstance(start, Distribution):
			raise ValueError("start must be a distribution.")

		if not isinstance(end, Distribution):
			raise ValueError("end must be a distribution.")

		if not isinstance(probability, float):
			raise ValueError("probability must be a float.")

		if self.edges is None:
			self.edges = []

		self.edges.append((start, end, probability))

	def probability(self, X, priors=None):
		"""Calculate the probability of each example.

		This method calculates the probability of each example given the
		parameters of the distribution. The examples must be given in a 3D
		format.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, length, self.d)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities).


		Returns
		-------
		prob: torch.Tensor, shape=(-1,)
			The probability of each example.
		"""

		return torch.exp(self.log_probability(X, priors=priors))

	def log_probability(self, X, priors=None):
		"""Calculate the log probability of each example.

		This method calculates the log probability of each example given the
		parameters of the distribution. The examples must be given in a 3D
		format.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, self.d)
			A set of examples to evaluate.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, length, self.d)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities).


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		f = self.forward(X, priors=priors)
		return torch.logsumexp(f[:, -1] + self.ends, dim=1)

	def predict_log_proba(self, X, priors=None):
		"""Calculate the posterior probabilities for each example.

		This method calculates the log posterior probabilities for each example
		and then normalizes across each component of the model. These
		probabilities are calculated using the forward-backward algorithm.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to summarize.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.n_distributions)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities).


		Returns
		-------
		r: torch.Tensor, shape=(-1, len, self.n_distributions)
			The log posterior probabilities for each example under each 
			component as calculated by the forward-backward algorithm.
		"""

		_, r, _, _, _ = self.forward_backward(X, priors=priors)
		return r

	def predict_proba(self, X, priors=None):
		"""Calculate the posterior probabilities for each example.

		This method calculates the posterior probabilities for each example
		and then normalizes across each component of the model. These
		probabilities are calculated using the forward-backward algorithm.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to summarize.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.n_distributions)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities).


		Returns
		-------
		y: torch.Tensor, shape=(-1, len, self.n_distributions)
			The posterior probabilities for each example under each component
			as calculated by the forward-backward algorithm.
		"""

		return torch.exp(self.predict_log_proba(X, priors=priors))

	def predict(self, X, priors=None):
		"""Predicts the component for each observation.

		This method calculates the predicted component for each observation
		given the posterior probabilities as calculated by the forward-backward
		algorithm. Essentially, it is just the argmax over components.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to summarize.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, length, self.d)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities).


		Returns
		-------
		y: torch.Tensor, shape=(-1, len, self.k)
			The posterior probabilities for each example under each component
			as calculated by the forward-backward algorithm.
		"""

		return torch.argmax(self.predict_log_proba(X, priors=priors), dim=-1)

	def fit(self, X, sample_weight=None, priors=None):
		"""Fit the model to sequences with optional weights and priors.

		This method implements the core of the learning process. For hidden
		Markov models, this is a form of EM called "Baum-Welch" or "structured
		EM". This iterative algorithm will proceed until converging, either
		according to the threshold set by `tol` or until the maximum number
		of iterations set by `max_iter` has been hit.

		This method is largely a wrapper around the `summarize` and
		`from_summaries` methods. It's primary contribution is serving as a
		loop around these functions and to monitor convergence.

		Unlike other HMM methods, this method can handle variable length
		sequences by accepting a list of tensors where each tensor has a
		different sequence length. Then, summarization is done on each tensor
		sequentially. This will provide an exact update as if the entire data
		set was seen at the same time but will allow batched operations to be
		performed on each variable length tensor.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to evaluate. Because sequences can be variable
			length, there are three ways to format the sequences.
			
				1. Pass in a tensor of shape (n, length, dim), which can only 
				be done when each sequence is the same length. 

				2. Pass in a list of 3D tensors where each tensor has the shape 
				(n, length, dim). In this case, each tensor is a collection of 
				sequences of the same length and so sequences of different 
				lengths can be trained on. 

				3. Pass in a list of 2D tensors where each tensor has the shape
				(length, dim). In this case, sequences of the same length will
				be grouped together into the same tensor and fitting will
				proceed as if you had passed in data like way 2.

		sample_weight: list, numpy.ndarray, torch.Tensor or None, optional
			A set of weights for the examples. These must follow the same format
			as X.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observations
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Must be formatted in the same
			shape as X. Default is None.


		Returns
		-------
		self
		"""

		X, sample_weight, priors = partition_sequences(X, 
			sample_weight=sample_weight, priors=priors, n_dists=self.k)

		# Initialize by concatenating across sequences
		if not self._initialized:
			if sample_weight is None:
				self._initialize(X)
			else:
				self._initialize(X, sample_weight=sample_weight)

		logp, last_logp = None, None
		for i in range(self.max_iter):
			start_time = time.time()

			# Train loop across all tensors
			logp = 0
			for j, X_ in enumerate(X):
				w_ = None if sample_weight is None else sample_weight[j]
				p_ = None if priors is None else priors[j]

				logp += self.summarize(X_, sample_weight=w_, priors=p_).sum()

			# Calculate and check improvement and optionally print it
			if i > 0:
				improvement = logp - last_logp
				duration = time.time() - start_time

				if self.verbose:
					print("[{}] Improvement: {}, Time: {:4.4}s".format(i, 
						improvement, duration))

				if improvement < self.tol:
					self._reset_cache()
					return self

			last_logp = logp
			self.from_summaries()

		# Calculate for the last iteration
		if self.verbose:
			logp = 0
			for j, X_ in enumerate(X):
				w_ = None if sample_weight is None else sample_weight[j]
				p_ = None if priors is None else priors[j]
				
				logp += self.summarize(X_, sample_weight=w_, priors=p_).sum()

			improvement = logp - last_logp
			duration = time.time() - start_time

			print("[{}] Improvement: {}, Time: {:4.4}s".format(i+1, 
				improvement, duration))

		self._reset_cache()
		return self

	def summarize(self, X, sample_weight=None, emissions=None, 
		priors=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: list, tuple, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d)
			A set of examples to summarize.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, length, self.d) or a vector of shape (-1,). Default is ones.

		emissions: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, n_distributions)
			Precalculated emission log probabilities. These are the
			probabilities of each observation under each probability 
			distribution. When running some algorithms it is more efficient
			to precalculate these and pass them into each call.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.n_distributions)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities).


		Returns
		-------
		logp: torch.Tensor, shape=(-1,)
			The log probability of each example.
		"""

		X = _check_parameter(_cast_as_tensor(X), "X", ndim=3, 
			shape=(-1, -1, self.d), check_parameter=self.check_data)
		emissions = _check_inputs(self, X, emissions, priors)

		if sample_weight is None:
			sample_weight = torch.ones(1, device=self.device).expand(
				emissions.shape[0], 1)
		else:
			sample_weight = _check_parameter(_cast_as_tensor(sample_weight),
				"sample_weight", min_value=0., ndim=1, 
				shape=(emissions.shape[0],), 
				check_parameter=self.check_data).reshape(-1, 1)

		if not self._initialized:
			self._initialize(X, sample_weight=sample_weight)

		return X, emissions, sample_weight

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		self.from_summaries()
		self._reset_cache()





# test_bayesian_network_structure_learning.py
# Contact: Jacob Schreiber <jmschreiber91@gmail.com>

import numpy
import torch
import pytest

from pomegranate.bayesian_network import BayesianNetwork

from pomegranate.bayesian_network import _from_structure
from pomegranate.bayesian_network import _learn_structure
from pomegranate.bayesian_network import _categorical_exact
from pomegranate.bayesian_network import _categorical_chow_liu

from pomegranate.distributions import Categorical
from pomegranate.distributions import ConditionalCategorical

from .tools import assert_raises
from numpy.testing import assert_array_almost_equal


def assert_tuple_equal(x, y):
	assert len(x) == len(y)

	for i in range(len(x)):
		assert len(x[i]) == len(y[i])

		if len(x[i]) > 0:
			for j in range(len(x[i])):
				assert x[i][j] == y[i][j]


@pytest.fixture
def X():
	return [[1, 2, 0, 0],
	     [0, 0, 1, 0],
	     [1, 1, 1, 0],
	     [1, 2, 1, 1],
	     [1, 1, 0, 1],
	     [0, 1, 0, 1],
	     [0, 1, 0, 0],
	     [1, 0, 1, 1],
	     [1, 1, 0, 1],
	     [0, 2, 1, 0],
	     [0, 0, 0, 1]]


@pytest.fixture
def X_masked(X):
	mask = torch.tensor(numpy.array([
		[False, True,  True,  True ],
		[True,  True,  False, False],
		[False, False, False, True ],
		[True,  True,  True,  True ],
		[False, True,  False, False],
		[True,  True,  True,  False],
		[False, False, False, True ],
		[True,  False, True,  True ],
		[True,  True,  True,  False],
		[True,  True,  True,  True ],
		[True,  False, True,  False]]))

	X = torch.tensor(numpy.array(X))
	return torch.masked.MaskedTensor(X, mask=mask)


@pytest.fixture
def w():
	return torch.tensor([1.0, 0.3, 1.0, 0.0, 0.0, 2.1, 3.1, 1.0, 1.2, 1.1, 2.0])


###


def test_categorical_chow_liu(X):
	structure = _categorical_chow_liu(X)
	assert_tuple_equal(structure, ((), (0,), (0,), (0,)))

	structure = _categorical_chow_liu(X, root=1)
	assert_tuple_equal(structure, ((1,), (), (1,), (1,)))

	structure = _categorical_chow_liu(X, root=2)
	assert_tuple_equal(structure, ((2,), (2,), (), (2,)))


def test_categorical_chow_liu_weighted(X, w):
	structure = _categorical_chow_liu(X, w)
	assert_tuple_equal(structure, ((), (0,), (0,), (0,)))

	structure = _categorical_chow_liu(X, w, root=1)
	assert_tuple_equal(structure, ((1,), (), (1,), (1,)))

	structure = _categorical_chow_liu(X, w, root=2)
	assert_tuple_equal(structure, ((2,), (2,), (), (2,)))


def test_categorical_chow_liu_large():
	numpy.random.seed(0)
	X = numpy.random.randint(3, size=(50, 10))

	structure = _categorical_chow_liu(X)
	assert_tuple_equal(structure, ((), (0,), (1,), (7,), (0,), (0,), (7,), (2,),
		(7,), (0,)))

	structure = _categorical_chow_liu(X, root=1)
	assert_tuple_equal(structure, ((1,), (), (1,), (7,), (0,), (0,), (7,), (2,), 
		(7,), (0,)))

	structure = _categorical_chow_liu(X, root=2)
	assert_tuple_equal(structure, ((1,), (2,), (), (7,), (0,), (0,), (7,), (2,), 
		(7,), (0,)))


def test_categorical_chow_liu_large_pseudocount():
	numpy.random.seed(0)
	X = numpy.random.randint(3, size=(50, 10))

	structure = _categorical_chow_liu(X, pseudocount=10)
	assert_tuple_equal(structure, ((), (2,), (4,), (2,), (0,), (3,), (3,), (2,), 
		(2,), (6,)))

	structure = _categorical_chow_liu(X, root=1, pseudocount=10)
	assert_tuple_equal(structure, ((4,), (), (1,), (2,), (2,), (3,), (3,), (2,), 
		(2,), (6,)))

	structure = _categorical_chow_liu(X, root=2, pseudocount=10)
	assert_tuple_equal(structure, ((4,), (2,), (), (2,), (2,), (3,), (3,), (2,),
		(2,), (6,)))


def test_categorical_chow_liu_raises(X, w):
	assert_raises(ValueError, _categorical_chow_liu, X, w, -1)
	assert_raises(TypeError, _categorical_chow_liu, X, w, [1.2, 1.1])
	assert_raises(ValueError, _categorical_chow_liu, X, w, None, -1)
	assert_raises(ValueError, _categorical_chow_liu, X, w, None, 0.3)
	assert_raises(ValueError, _categorical_chow_liu, X, w, None, [1, 2])

	w = torch.tensor(w)
	assert_raises(ValueError, _categorical_chow_liu, X, w.unsqueeze(0))
	assert_raises(ValueError, _categorical_chow_liu, X, w.unsqueeze(1))

	X = numpy.array(X)
	n, d = X.shape

	assert_raises(ValueError, _categorical_chow_liu, X + 0.3, w)
	assert_raises(ValueError, _categorical_chow_liu, X.reshape(1, n, d), w)
	assert_raises(ValueError, _categorical_chow_liu, X - 1, w)


###


def test_categorical_exact(X):
	structure = _categorical_exact(X)
	assert_tuple_equal(structure, ((), (0,), (0, 1), ()))

	structure = _categorical_exact(X, max_parents=1)
	assert_tuple_equal(structure, ((), (0,), (), ()))


def test_categorical_exact_weighted(X, w):
	structure = _categorical_exact(X, w)
	assert_tuple_equal(structure, ((), (0,), (0, 1), ()))

	structure = _categorical_exact(X, w, max_parents=1)
	assert_tuple_equal(structure, ((), (0,), (), ()))


def test_categorical_exact_exclude_parents(X):
	exclude_parents = ((), (2,), (), (1,))
	structure = _categorical_exact(X, exclude_parents=exclude_parents)
	assert_tuple_equal(structure, ((), (), (0,), (0, 2)))

	structure = _categorical_exact(X, exclude_parents=exclude_parents, 
		max_parents=1)
	assert_tuple_equal(structure, ((), (0,), (), ()))

	exclude_parents = ((), (2,), (), (0, 1))
	structure = _categorical_exact(X, exclude_parents=exclude_parents)
	assert_tuple_equal(structure, ((2, 3), (), (), (2,)))


def test_categorical_exact_large():
	numpy.random.seed(0)
	X = numpy.random.randint(3, size=(50, 8))

	structure = _categorical_exact(X)
	assert_tuple_equal(structure, ((), (0, 3), (0, 1, 3), (), (0, 1, 2, 3), (), 
		(), ()))

	structure = _categorical_exact(X, max_parents=1)
	assert_tuple_equal(structure, ((), (), (), (), (), (0,), (), ()))

	structure = _categorical_exact(X, max_parents=2)
	assert_tuple_equal(structure, ((), (), (0, 1), (), (), (), (), ()))

	structure = _categorical_exact(X, max_parents=3)
	assert_tuple_equal(structure, ((), (), (0, 1), (), (0, 1, 2), (), (), ()))

	structure = _categorical_exact(X, max_parents=4)
	assert_tuple_equal(structure, ((), (0, 3), (0, 1, 3), (), (0, 1, 2, 3), (), 
		(), ()))


def test_categorical_exact_large_pseudocount():
	numpy.random.seed(0)
	X = numpy.random.randint(3, size=(50, 8))

	structure = _categorical_exact(X, pseudocount=10)
	assert_tuple_equal(structure, ((), (), (), (), (), (), (), ()))

	structure = _categorical_exact(X, pseudocount=5)
	assert_tuple_equal(structure, ((), (), (), (), (), (), (), ()))

	structure = _categorical_exact(X, pseudocount=2)
	assert_tuple_equal(structure, ((), (), (), (), (), (), (), ()))

	structure = _categorical_exact(X, pseudocount=1)
	assert_tuple_equal(structure, ((), (), (), (), (), (), (), ()))

	structure = _categorical_exact(X, pseudocount=0.1)
	assert_tuple_equal(structure, ((), (), (), (), (), (), (), ()))

	structure = _categorical_exact(X, pseudocount=1e-8)
	assert_tuple_equal(structure, ((), (), (), (), (), (0,), (), ()))


###


def test_categorical_learn_structure_chow_liu():
	numpy.random.seed(0)
	X = numpy.random.randint(3, size=(50, 7))
	w = numpy.abs(numpy.random.randn(50))

	structure1 = _categorical_chow_liu(X)
	structure2 = _learn_structure(X, algorithm='chow-liu')
	assert_tuple_equal(structure1, structure2)

	structure1 = _categorical_chow_liu(X, root=1)
	structure2 = _learn_structure(X, root=1, algorithm='chow-liu')
	assert_tuple_equal(structure1, structure2)

	structure1 = _categorical_chow_liu(X, pseudocount=50)
	structure2 = _learn_structure(X, pseudocount=50, algorithm='chow-liu')
	assert_tuple_equal(structure1, structure2)

	structure1 = _categorical_chow_liu(X, w)
	structure2 = _learn_structure(X, w, algorithm='chow-liu')
	assert_tuple_equal(structure1, structure2)


def test_categorical_learn_structure_exact():
	numpy.random.seed(0)
	X = numpy.random.randint(3, size=(50, 7))
	w = numpy.abs(numpy.random.randn(50))

	structure1 = _categorical_exact(X)
	structure2 = _learn_structure(X, algorithm='exact')
	assert_tuple_equal(structure1, structure2)

	structure1 = _categorical_exact(X, max_parents=1)
	structure2 = _learn_structure(X, max_parents=1, algorithm='exact')
	assert_tuple_equal(structure1, structure2)

	structure1 = _categorical_exact(X, pseudocount=50)
	structure2 = _learn_structure(X, pseudocount=50, algorithm='exact')
	assert_tuple_equal(structure1, structure2)

	structure1 = _categorical_exact(X, w)
	structure2 = _learn_structure(X, w, algorithm='exact')
	assert_tuple_equal(structure1, structure2)


###


def test_categorical_from_structure(X):
	structure = ((), (0,), (1, 3), ())
	distributions = _from_structure(X, structure=structure)
	model = BayesianNetwork(distributions, structure=structure)

	assert isinstance(model.distributions[0], Categorical)
	assert isinstance(model.distributions[1], ConditionalCategorical)
	assert isinstance(model.distributions[2], ConditionalCategorical)
	assert isinstance(model.distributions[3], Categorical)

	p0 = model.distributions[0].probs
	p1 = model.distributions[1].probs[0]
	p2 = model.distributions[2].probs[0]
	p3 = model.distributions[3].probs

	assert_array_almost_equal(p0, [[0.454545, 0.545455]])
	assert_array_almost_equal(p1, [[0.4     , 0.4     , 0.2     ], 
		[0.166667, 0.5     , 0.333333]])
	assert_array_almost_equal(p2, [[[0.0000, 1.0000],
         [0.5000, 0.5000]],

        [[0.5000, 0.5000],
         [1.0000, 0.0000]],

        [[0.5000, 0.5000],
         [0.0000, 1.0000]]])
	assert_array_almost_equal(p3, [[0.4545, 0.5455]], 4)


def test_categorical_from_structure_weighted(X, w):
	structure = ((), (0,), (1, 3), ())
	distributions = _from_structure(X, sample_weight=w, structure=structure)
	model = BayesianNetwork(distributions, structure=structure)

	assert isinstance(model.distributions[0], Categorical)
	assert isinstance(model.distributions[1], ConditionalCategorical)
	assert isinstance(model.distributions[2], ConditionalCategorical)
	assert isinstance(model.distributions[3], Categorical)

	p0 = model.distributions[0].probs
	p1 = model.distributions[1].probs[0]
	p2 = model.distributions[2].probs[0]
	p3 = model.distributions[3].probs

	assert_array_almost_equal(p0, [[0.6719, 0.3281]], 4)
	assert_array_almost_equal(p1, [[0.2674, 0.6047, 0.1279], 
		[0.2381, 0.5238, 0.2381]], 4)
	assert_array_almost_equal(p2, [[[0.0000, 1.0000],
         [0.6667, 0.3333]],

        [[0.7561, 0.2439],
         [1.0000, 0.0000]],

        [[0.4762, 0.5238],
         [0.5000, 0.5000]]], 4)
	assert_array_almost_equal(p3, [[0.5078, 0.4922]], 4)


def test_categorical_from_structure_null(X):
	structure = ((), (), (), ())
	distributions = _from_structure(X, structure=structure)
	model = BayesianNetwork(distributions, structure=structure)

	assert isinstance(model.distributions[0], Categorical)
	assert isinstance(model.distributions[1], Categorical)
	assert isinstance(model.distributions[2], Categorical)
	assert isinstance(model.distributions[3], Categorical)

	p0 = model.distributions[0].probs
	p1 = model.distributions[1].probs
	p2 = model.distributions[2].probs
	p3 = model.distributions[3].probs

	assert_array_almost_equal(p0, [[0.454545, 0.545455]])
	assert_array_almost_equal(p1, [[0.272727, 0.454545, 0.272727]])
	assert_array_almost_equal(p2, [[0.5455, 0.4545]], 4)
	assert_array_almost_equal(p3, [[0.4545, 0.5455]], 4)



# dense_hmm.py
# Author: Jacob Schreiber <jmschreiber91@gmail.com>

import math
import numpy
import torch

from .._utils import _cast_as_tensor
from .._utils import _cast_as_parameter
from .._utils import _update_parameter
from .._utils import _check_parameter

from ..distributions._distribution import Distribution

from ._base import _BaseHMM
from ._base import _check_inputs


NEGINF = float("-inf")
inf = float("inf")


class DenseHMM(_BaseHMM):
	"""A hidden Markov model with a dense transition matrix.

	A hidden Markov model is an extension of a mixture model to sequences by
	including a transition matrix between the elements of the mixture. Each of
	the algorithms for a hidden Markov model are essentially just a revision
	of those algorithms to incorporate this transition matrix.

	This object is a wrapper for a hidden Markov model with a dense transition
	matrix.

	This object is a wrapper for both implementations, which can be specified
	using the `kind` parameter. Choosing the right implementation will not
	effect the accuracy of the results but will change the speed at which they
	are calculated. 	

	Separately, there are two ways to instantiate the hidden Markov model. The
	first is by passing in a set of distributions, a dense transition matrix, 
	and optionally start/end probabilities. The second is to initialize the
	object without these and then to add edges using the `add_edge` method
	and to add distributions using the `add_distributions` method. Importantly, the way that
	you choose to initialize the hidden Markov model is independent of the
	implementation that you end up choosing. If you pass in a dense transition
	matrix, this will be converted to a sparse matrix with all the zeros
	dropped if you choose `kind='sparse'`.


	Parameters
	----------
	distributions: tuple or list
		A set of distribution objects. These objects do not need to be
		initialized, i.e., can be "Normal()". 

	edges: numpy.ndarray, torch.Tensor, or None. shape=(k,k), optional
		A dense transition matrix of probabilities for how each node or
		distribution passed in connects to each other one. This can contain
		many zeroes, and when paired with `kind='sparse'`, will drop those
		elements from the matrix.

	starts: list, numpy.ndarray, torch.Tensor, or None. shape=(k,), optional
		The probability of starting at each node. If not provided, assumes
		these probabilities are uniform.

	ends: list, numpy.ndarray, torch.Tensor, or None. shape=(k,), optional
		The probability of ending at each node. If not provided, assumes
		these probabilities are uniform.

	inertia: float, [0, 1], optional
		Indicates the proportion of the update to apply to the parameters
		during training. When the inertia is 0.0, the update is applied in
		its entirety and the previous parameters are ignored. When the
		inertia is 1.0, the update is entirely ignored and the previous
		parameters are kept, equivalently to if the parameters were frozen.

	frozen: bool, optional
		Whether all the parameters associated with this distribution are frozen.
		If you want to freeze individual pameters, or individual values in those
		parameters, you must modify the `frozen` attribute of the tensor or
		parameter directly. Default is False.

	check_data: bool, optional
		Whether to check properties of the data and potentially recast it to
		torch.tensors. This does not prevent checking of parameters but can
		slightly speed up computation when you know that your inputs are valid.
		Setting this to False is also necessary for compiling. Default is True.
	"""

	def __init__(self, distributions=None, edges=None, starts=None, ends=None, 
		init='random', max_iter=1000, tol=0.1, sample_length=None, 
		return_sample_paths=False, inertia=0.0, frozen=False, check_data=True,
		random_state=None, verbose=False):
		super().__init__(distributions=distributions, starts=starts, ends=ends, 
			init=init, max_iter=max_iter, tol=tol, sample_length=sample_length, 
			return_sample_paths=return_sample_paths, inertia=inertia,
			frozen=frozen, check_data=check_data, random_state=random_state, 
			verbose=verbose)
		
		self.name = "DenseHMM"
		n = len(distributions) if distributions is not None else 0

		if edges is not None:
			self.edges = _cast_as_parameter(torch.log(_check_parameter(
				_cast_as_tensor(edges), "edges", ndim=2, shape=(n, n), 
				min_value=0., max_value=1.)))

		self._initialized = (self.distributions is not None and
			self.starts is not None and self.ends is not None and
			self.edges is not None and 
			all(d._initialized for d in self.distributions))

		if self._initialized:
			self.distributions = torch.nn.ModuleList(
				self.distributions)

		self._reset_cache()

	def _reset_cache(self):
		"""Reset the internally stored statistics.

		This method is meant to only be called internally. It resets the
		stored statistics used to update the model parameters as well as
		recalculates the cached values meant to speed up log probability
		calculations.
		"""

		if self._initialized == False:
			return

		for node in self.distributions:
			node._reset_cache()

		self.register_buffer("_xw_sum", torch.zeros(self.n_distributions, 
			self.n_distributions, dtype=self.dtype, requires_grad=False, 
			device=self.device))

		self.register_buffer("_xw_starts_sum", torch.zeros(self.n_distributions, 
			dtype=self.dtype, requires_grad=False, device=self.device))

		self.register_buffer("_xw_ends_sum", torch.zeros(self.n_distributions, 
			dtype=self.dtype, requires_grad=False, device=self.device))

	def _initialize(self, X=None, sample_weight=None):
		"""Initialize the probability distribution.

		This method is meant to only be called internally. It initializes the
		parameters of the distribution and stores its dimensionality. For more
		complex methods, this function will do more.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, len, self.d), optional
			The data to use to initialize the model. Default is None.

		sample_weight: list, tuple, numpy.ndarray, torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, len) or a vector of shape (-1,). If None, defaults to ones.
			Default is None.
		"""

		super()._initialize(X, sample_weight=sample_weight)

		n = self.n_distributions
		if self.edges == None:
			self.edges = _cast_as_parameter(torch.log(torch.ones(n, n, 
				dtype=self.dtype, device=self.device) / n))

		self.distributions = torch.nn.ModuleList(self.distributions)

	def add_edge(self, start, end, prob):
		"""Add an edge to the model.

		This method will fill in an entry in the dense transition matrix
		at row indexed by the start distribution and the column indexed
		by the end distribution. The value that will be included is the
		log of the probability value provided. Note that this will override
		values that already exist, and that this will initialize a new
		dense transition matrix if none has been passed in so far.


		Parameters
		----------
		start: torch.distributions.distribution
			The distribution that the edge starts at.

		end: torch.distributions.distribution
			The distribution that the edge ends at.

		prob: float, (0.0, 1.0]
			The probability of that edge.
		"""

		if self.distributions is None:
			raise ValueError("Must add distributions before edges.")

		n = self.n_distributions

		if start == self.start:
			if self.starts is None:
				self.starts = torch.empty(n, dtype=self.dtype, 
					device=self.device) - inf

			idx = self.distributions.index(end)
			self.starts[idx] = math.log(prob)

		elif end == self.end:
			if self.ends is None:
				self.ends = torch.empty(n, dtype=self.dtype, 
					device=self.device) - inf

			idx = self.distributions.index(start)
			self.ends[idx] = math.log(prob)

		else:
			if self.edges is None:
				self.edges = torch.empty((n, n), dtype=self.dtype, 
					device=self.device) - inf

			idx1 = self.distributions.index(start)
			idx2 = self.distributions.index(end)

			self.edges[idx1, idx2] = math.log(prob)

	def sample(self, n):
		"""Sample from the probability distribution.

		This method will return `n` samples generated from the underlying
		probability distribution. Because a HMM describes variable length
		sequences, a list will be returned where each element is one of
		the generated sequences.


		Parameters
		----------
		n: int
			The number of samples to generate.
		

		Returns
		-------
		X: list of torch.tensor, shape=(n,)
			A list of randomly generated samples, where each sample of
			size (length, self.d).
		"""

		if self.sample_length is None and self.ends is None:
			raise ValueError("Must specify a length or have explicit "
				+ "end probabilities.")

		if self.ends is None:
			ends = torch.zeros(self.n_distributions, dtype=self.edges.dtype, 
				device=self.edges.device) + float("-inf")
		else:
			ends = self.ends

		distributions, emissions = [], []

		edge_probs = torch.hstack([self.edges, ends.unsqueeze(1)])
		edge_probs = torch.exp(edge_probs).numpy()

		starts = torch.exp(self.starts).numpy()

		for _ in range(n):
			node_i = self.random_state.choice(self.n_distributions, p=starts)
			emission_i = self.distributions[node_i].sample(n=1)
			distributions_, emissions_ = [node_i], [emission_i]

			for i in range(1, self.sample_length or int(1e8)):
				node_i = self.random_state.choice(self.n_distributions+1, 
					p=edge_probs[node_i])

				if node_i == self.n_distributions:
					break

				emission_i = self.distributions[node_i].sample(n=1)

				distributions_.append(node_i)
				emissions_.append(emission_i)

			distributions.append(distributions_)
			emissions.append(torch.vstack(emissions_))

		if self.return_sample_paths == True:
			return emissions, distributions
		return emissions

	def viterbi(self, X=None, emissions=None, priors=None):
		"""Run the Viterbi algorithm on some data.

		Runs the Viterbi algortihm on a batch of sequences. The Viterbi 
		algorithm is a dynamic programming algorithm that begins at the start
		state and calculates the single best path through the model involving
		alignments of symbol i to node j. This is in contrast to the forward
		function, which involves calculating the sum of all paths, not just
		the single best path. Because we have to keep track of the best path,
		the Viterbi algorithm is slightly more conceptually challenging and
		involves keeping track of a traceback matrix.


		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, d)
			A set of examples to evaluate. Does not need to be passed in if
			emissions are. 

		emissions: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, n_dists)
			Precalculated emission log probabilities. These are the
			probabilities of each observation under each probability 
			distribution. When running some algorithms it is more efficient
			to precalculate these and pass them into each call.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		path: torch.Tensor, shape=(-1, -1)
			The state assignment for each observation in each sequence.
		""" 

		emissions = _check_inputs(self, X, emissions, priors)
		n, l = emissions.shape[:2]

		v = torch.clone(emissions.permute(1, 0, 2)).contiguous()
		v[0] += self.starts
		
		traceback = torch.zeros_like(v, dtype=torch.int32)
		traceback[0] = torch.arange(v.shape[-1])

		for i in range(1, l):
			z = v[i-1].unsqueeze(-1) + self.edges.unsqueeze(0) + v[i][:, None]
			v[i], traceback[i] = torch.max(z, dim=-2)

		ends = self.ends + v[-1]
		best_end_logps, best_end_idxs = torch.max(ends, dim=-1)

		paths = [best_end_idxs]
		for i in range(1, l):
			paths.append(traceback[l-i, torch.arange(n), paths[-1]])

		paths = torch.flip(torch.stack(paths).T, dims=(-1,))
		return paths


	def forward(self, X=None, emissions=None, priors=None):
		"""Run the forward algorithm on some data.

		Runs the forward algorithm on a batch of sequences. This is not to be
		confused with a "forward pass" when talking about neural networks. The
		forward algorithm is a dynamic programming algorithm that begins at the
		start state and returns the probability, over all paths through the
		model, that result in the alignment of symbol i to node j.

		
		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, d)
			A set of examples to evaluate. Does not need to be passed in if
			emissions are. 

		emissions: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, n_dists)
			Precalculated emission log probabilities. These are the
			probabilities of each observation under each probability 
			distribution. When running some algorithms it is more efficient
			to precalculate these and pass them into each call.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		f: torch.Tensor, shape=(-1, -1, self.n_distributions)
			The log probabilities calculated by the forward algorithm.
		"""
		
		emissions = _check_inputs(self, X, emissions, priors)
		l = emissions.shape[1]

		t_max = self.edges.max()
		t = torch.exp(self.edges - t_max)
		f = torch.clone(emissions.permute(1, 0, 2)).contiguous()
		f[0] += self.starts
		f[1:] += t_max

		for i in range(1, l):
			p_max = torch.max(f[i-1], dim=1, keepdims=True).values
			p = torch.exp(f[i-1] - p_max)
			f[i] += torch.log(torch.matmul(p, t)) + p_max

		f = f.permute(1, 0, 2)
		return f

	def backward(self, X=None, emissions=None, priors=None):
		"""Run the backward algorithm on some data.

		Runs the backward algorithm on a batch of sequences. This is not to be
		confused with a "backward pass" when talking about neural networks. The
		backward algorithm is a dynamic programming algorithm that begins at end
		of the sequence and returns the probability, over all paths through the
		model, that result in the alignment of symbol i to node j, working
		backwards.

		
		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, d)
			A set of examples to evaluate. Does not need to be passed in if
			emissions are. 

		emissions: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, n_distributions)
			Precalculated emission log probabilities. These are the
			probabilities of each observation under each probability 
			distribution. When running some algorithms it is more efficient
			to precalculate these and pass them into each call.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		b: torch.Tensor, shape=(-1, length, self.n_distributions)
			The log probabilities calculated by the backward algorithm.
		"""

		emissions = _check_inputs(self, X, emissions, priors)
		n, l, _ = emissions.shape

		b = torch.zeros(l, n, self.n_distributions, dtype=self.dtype, 
			device=self.device) + float("-inf")
		b[-1] = self.ends

		t_max = self.edges.max()
		t = torch.exp(self.edges.T - t_max)

		for i in range(l-2, -1, -1):
			p = b[i+1] + emissions[:, i+1]
			p_max = torch.max(p, dim=1, keepdims=True).values
			p = torch.exp(p - p_max)

			b[i] = torch.log(torch.matmul(p, t)) + t_max + p_max

		b = b.permute(1, 0, 2)
		return b

	def forward_backward(self, X=None, emissions=None, priors=None):
		"""Run the forward-backward algorithm on some data.

		Runs the forward-backward algorithm on a batch of sequences. This
		algorithm combines the best of the forward and the backward algorithm.
		It combines the probability of starting at the beginning of the sequence
		and working your way to each observation with the probability of
		starting at the end of the sequence and working your way backward to it.

		A number of statistics can be calculated using this information. These
		statistics are powerful inference tools but are also used during the
		Baum-Welch training process. 

		
		Parameters
		----------
		X: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, d)
			A set of examples to evaluate. Does not need to be passed in if
			emissions are. 

		emissions: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, n_distributions)
			Precalculated emission log probabilities. These are the
			probabilities of each observation under each probability 
			distribution. When running some algorithms it is more efficient
			to precalculate these and pass them into each call.

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.


		Returns
		-------
		transitions: torch.Tensor, shape=(-1, n, n)
			The expected number of transitions across each edge that occur
			for each example. The returned transitions follow the structure
			of the transition matrix and so will be dense or sparse as
			appropriate.

		responsibility: torch.Tensor, shape=(-1, -1, n)
			The posterior probabilities of each observation belonging to each
			state given that one starts at the beginning of the sequence,
			aligns observations across all paths to get to the current
			observation, and then proceeds to align all remaining observations
			until the end of the sequence.

		starts: torch.Tensor, shape=(-1, n)
			The probabilities of starting at each node given the 
			forward-backward algorithm.

		ends: torch.Tensor, shape=(-1, n)
			The probabilities of ending at each node given the forward-backward
			algorithm.

		logp: torch.Tensor, shape=(-1,)
			The log probabilities of each sequence given the model.
		"""

		emissions = _check_inputs(self, X, emissions, priors)
		n, l, _ = emissions.shape

		f = self.forward(emissions=emissions)
		b = self.backward(emissions=emissions)

		logp = torch.logsumexp(f[:, -1] + self.ends, dim=1)

		f_ = f[:, :-1].unsqueeze(-1)
		b_ = (b[:, 1:] + emissions[:, 1:]).unsqueeze(-2)

		t = f_ + b_ + self.edges.unsqueeze(0).unsqueeze(0)
		t = t.reshape(n, l-1, -1)
		t = torch.exp(torch.logsumexp(t, dim=1).T - logp).T
		t = t.reshape(n, int(t.shape[1] ** 0.5), -1)

		starts = self.starts + emissions[:, 0] + b[:, 0]
		starts = torch.exp(starts.T - torch.logsumexp(starts, dim=-1)).T

		ends = self.ends + f[:, -1]
		ends = torch.exp(ends.T - torch.logsumexp(ends, dim=-1)).T

		r = f + b
		r = r - torch.logsumexp(r, dim=2).reshape(n, -1, 1)
		return t, r, starts, ends, logp

	def summarize(self, X, sample_weight=None, emissions=None, priors=None):
		"""Extract the sufficient statistics from a batch of data.

		This method calculates the sufficient statistics from optionally
		weighted data and adds them to the stored cache. The examples must be
		given in a 2D format. Sample weights can either be provided as one
		value per example or as a 2D matrix of weights for each feature in
		each example.


		Parameters
		----------
		X: torch.Tensor, shape=(-1, -1, self.d)
			A set of examples to summarize.

		y: torch.Tensor, shape=(-1, -1), optional 
			A set of labels with the same number of examples and length as the
			observations that indicate which node in the model that each
			observation should be assigned to. Passing this in means that the
			model uses labeled training instead of Baum-Welch. Default is None.

		sample_weight: torch.Tensor, optional
			A set of weights for the examples. This can be either of shape
			(-1, self.d) or a vector of shape (-1,). Default is ones.

		emissions: torch.Tensor, shape=(-1, -1, self.n_distributions)
			Precalculated emission log probabilities. These are the
			probabilities of each observation under each probability 
			distribution. When running some algorithms it is more efficient
			to precalculate these and pass them into each call.	

		priors: list, numpy.ndarray, torch.Tensor, shape=(-1, -1, self.k)
			Prior probabilities of assigning each symbol to each node. If not
			provided, do not include in the calculations (conceptually
			equivalent to a uniform probability, but without scaling the
			probabilities). This can be used to assign labels to observatons
			by setting one of the probabilities for an observation to 1.0.
			Note that this can be used to assign hard labels, but does not
			have the same semantics for soft labels, in that it only
			influences the initial estimate of an observation being generated
			by a component, not gives a target. Default is None.
		"""

		X, emissions, sample_weight = super().summarize(X, 
			sample_weight=sample_weight, emissions=emissions, priors=priors)

		t, r, starts, ends, logps = self.forward_backward(emissions=emissions)

		X = X.reshape(-1, X.shape[-1])
		r = torch.exp(r) * sample_weight.unsqueeze(-1)
		for i, node in enumerate(self.distributions):
			w = r[:, :, i].reshape(-1, 1)
			node.summarize(X, sample_weight=w)

		if self.frozen == False:
			self._xw_starts_sum += torch.sum(starts * sample_weight, dim=0)
			self._xw_ends_sum += torch.sum(ends * sample_weight, dim=0)
			self._xw_sum += torch.sum(t * sample_weight.unsqueeze(-1), dim=0) 

		return logps

	def from_summaries(self):
		"""Update the model parameters given the extracted statistics.

		This method uses calculated statistics from calls to the `summarize`
		method to update the distribution parameters. Hyperparameters for the
		update are passed in at initialization time.

		Note: Internally, a call to `fit` is just a successive call to the
		`summarize` method followed by the `from_summaries` method.
		"""

		for node in self.distributions:
			node.from_summaries()

		if self.frozen:
			return

		node_out_count = torch.sum(self._xw_sum, dim=1, keepdims=True)
		node_out_count += self._xw_ends_sum.unsqueeze(1)

		ends = torch.log(self._xw_ends_sum / node_out_count[:,0])
		starts = torch.log(self._xw_starts_sum / self._xw_starts_sum.sum())
		edges = torch.log(self._xw_sum / node_out_count)

		_update_parameter(self.ends, ends, inertia=self.inertia)
		_update_parameter(self.starts, starts, inertia=self.inertia)
		_update_parameter(self.edges, edges, inertia=self.inertia)
		self._reset_cache()
